@@ -16,7 +16,7 @@ use bitflags::bitflags;
 use libc::{c_char, c_int};
 
 use std::cell::RefCell;
-use std::ffi::{CStr, CString};
+use std::ffi::{c_void, CStr, CString};
 use std::fmt::{self, Display, Formatter};
 use std::marker::{self, PhantomData};
 use std::mem;
@@ -223,10 +223,11 @@ macro_rules! impl_wired_simple {
 
             fn serialize(&self, wire_buf: &mut [u8]) -> usize {
                 debug_assert!(wire_buf.len() >= mem::size_of::<Self>());
-                unsafe {
-                    *(wire_buf.as_mut_ptr() as *mut Self) = *self;
-                }
-                mem::size_of::<Self>()
+                let buf_size = self.wire_len();
+                let src =
+                    unsafe { slice::from_raw_parts(self as *const Self as *const u8, buf_size) };
+                wire_buf[..buf_size].copy_from_slice(src);
+                buf_size
             }
         }
 
@@ -265,10 +266,12 @@ impl<T: Xid> WiredOut for T {
 
     fn serialize(&self, wire_buf: &mut [u8]) -> usize {
         debug_assert!(wire_buf.len() >= 4);
-        unsafe {
-            *(wire_buf.as_mut_ptr() as *mut u32) = self.resource_id();
-        }
-        4
+        let buf_size = self.wire_len();
+        let resource_id = self.resource_id();
+        let src =
+            unsafe { slice::from_raw_parts(&resource_id as *const u32 as *const u8, buf_size) };
+        wire_buf[..buf_size].copy_from_slice(src);
+        buf_size
     }
 }
 
@@ -423,6 +426,8 @@ pub trait RequestWithReply: Request {
 
 /// Determines whether Xlib or XCB owns the event queue of [`Connection`].
 ///
+/// This item is behind the `xlib_xcb` cargo feature.
+///
 /// See [`Connection::set_event_queue_owner`].
 #[cfg(feature = "xlib_xcb")]
 #[derive(Debug)]
@@ -533,6 +538,9 @@ pub enum ConnError {
     ClosedInvalidScreen,
     /// Connection closed because some file descriptor passing operation failed.
     ClosedFdPassingFailed,
+    /// XOpenDisplay returned NULL
+    #[cfg(feature = "xlib_xcb")]
+    XOpenDisplay,
 }
 
 impl ConnError {
@@ -551,6 +559,10 @@ impl ConnError {
             ConnError::ClosedFdPassingFailed => {
                 "Connection closed, some file descriptor passing operation failed"
             }
+            #[cfg(feature = "xlib_xcb")]
+            ConnError::XOpenDisplay => {
+                "XOpenDisplay failed to open a display. Check the $DISPLAY env var"
+            }
         }
     }
 }
@@ -714,10 +726,15 @@ impl Connection {
     /// The event queue owner defaults to Xlib.
     /// One would need to open an XCB connection with Xlib in order to use
     /// OpenGL.
+    ///
+    /// This function is behind the `xlib_xcb` cargo feature.
     #[cfg(feature = "xlib_xcb")]
     pub fn connect_with_xlib_display() -> ConnResult<(Connection, i32)> {
         unsafe {
             let dpy = xlib::XOpenDisplay(ptr::null());
+            if dpy.is_null() {
+                return Err(ConnError::XOpenDisplay);
+            }
 
             check_connection_error(XGetXCBConnection(dpy))?;
 
@@ -737,6 +754,8 @@ impl Connection {
     /// One would need to open an XCB connection with Xlib in order to use
     /// OpenGL.
     ///
+    /// This function is behind the `xlib_xcb` cargo feature.
+    ///
     /// # Panics
     /// Panics if one of the mandatory extension is not present.
     #[cfg(feature = "xlib_xcb")]
@@ -746,6 +765,9 @@ impl Connection {
     ) -> ConnResult<(Connection, i32)> {
         unsafe {
             let dpy = xlib::XOpenDisplay(ptr::null());
+            if dpy.is_null() {
+                return Err(ConnError::XOpenDisplay);
+            }
 
             check_connection_error(XGetXCBConnection(dpy))?;
 
@@ -948,6 +970,8 @@ impl Connection {
     /// Wraps a `xlib::Display` and get an XCB connection from an exisiting object
     /// `xlib::XCloseDisplay` will be called when the returned object is dropped.
     ///
+    /// This function is behind the `xlib_xcb` cargo feature.
+    ///
     /// # Safety
     /// The `dpy` pointer must be a pointer to a valid `xlib::Display`
     #[cfg(feature = "xlib_xcb")]
@@ -963,6 +987,8 @@ impl Connection {
     /// Extension data specified by `mandatory` and `optional` is cached to allow
     /// the resolution of events and errors in these extensions.
     ///
+    /// This function is behind the `xlib_xcb` cargo feature.
+    ///
     /// # Panics
     /// Panics if the connection is null or in error state.
     ///
@@ -1036,6 +1062,8 @@ impl Connection {
     }
 
     /// Returns the inner ffi `xlib::Display` pointer.
+    ///
+    /// This function is behind the `xlib_xcb` cargo feature.
     #[cfg(feature = "xlib_xcb")]
     pub fn get_raw_dpy(&self) -> *mut xlib::Display {
         self.dpy
@@ -1043,6 +1071,8 @@ impl Connection {
 
     /// Sets the owner of the event queue in the case if the connection is opened
     /// with the Xlib interface. In that case, the default owner is Xlib.
+    ///
+    /// This function is behind the `xlib_xcb` cargo feature.
     #[cfg(feature = "xlib_xcb")]
     pub fn set_event_queue_owner(&self, owner: EventQueueOwner) {
         debug_assert!(!self.dpy.is_null());
@@ -1114,6 +1144,12 @@ impl Connection {
     ///
     /// There are several occasions ones want to flush the connection.
     /// One of them is before entering or re-entering the event loop after performing unchecked requests.
+    ///
+    /// The main difference between `flush` and `check_request` is that `flush` will not report protocol errors.
+    /// If a protocol error is emitted by an unchecked void request, it will be reported through the event loop.
+    ///
+    /// See also: [wait_for_event](Connection::wait_for_event), [check_request](Connection::check_request),
+    /// [send_and_check_request](Connection::send_and_check_request).
     pub fn flush(&self) -> ConnResult<()> {
         unsafe {
             let ret = xcb_flush(self.c);
@@ -1126,26 +1162,14 @@ impl Connection {
         }
     }
 
-    unsafe fn handle_wait_for_event(&self, ev: *mut xcb_generic_event_t) -> Result<Event> {
-        if ev.is_null() {
-            self.has_error()?;
-            panic!("xcb_wait_for_event returned null with I/O error");
-        } else if is_error(ev) {
-            Err(error::resolve_error(ev as *mut _, &self.ext_data).into())
-        } else {
-            Ok(event::resolve_event(ev, &self.ext_data))
-        }
-    }
-
-    unsafe fn handle_poll_for_event(&self, ev: *mut xcb_generic_event_t) -> Result<Option<Event>> {
-        if ev.is_null() {
-            self.has_error()?;
-            Ok(None)
-        } else if is_error(ev) {
-            Err(error::resolve_error(ev as *mut _, &self.ext_data).into())
-        } else {
-            Ok(Some(event::resolve_event(ev, &self.ext_data)))
-        }
+    /// Resolve an xcb_generic_event_t pointer into an Event.
+    /// # Safety
+    /// The caller is repsonsible to ensure that the `ev` pointer is not NULL.
+    /// The ownership of the pointer is effectively transferred to the
+    /// returned Event and it will be destroyed when the Event is
+    /// dropped.
+    pub unsafe fn resolve_event(&self, ev: &mut xcb_generic_event_t) -> Event {
+        event::resolve_event(ev, &self.ext_data)
     }
 
     /// Blocks and returns the next event or error from the server.
@@ -1230,6 +1254,8 @@ impl Connection {
     ///
     /// Effectively creates an internal special queue for this event
     /// XGE events are only defined in the `xinput` and `present` extensions
+    ///
+    /// This function is present only if either of the `xinput` or `present` cargo features are active.
     #[cfg(any(feature = "xinput", feature = "present"))]
     pub fn register_for_special_xge<XGE: GeEvent>(&self) -> SpecialEventId {
         unsafe {
@@ -1322,7 +1348,8 @@ impl Connection {
     /// Send a request to the X server.
     ///
     /// This function never blocks. A cookie is returned to keep track of the request.
-    /// If the request expect a reply, the cookie can be used to retrieve the reply
+    /// If the request expect a reply, the cookie can be used to retrieve the reply with
+    /// [Connection::wait_for_reply].
     ///
     /// # Example
     /// ```no_run
@@ -1458,7 +1485,36 @@ impl Connection {
         }
     }
 
-    /// Get the reply of a previous request, or an error if one occured.
+    /// Send the request to the server and check it.
+    ///
+    /// This is a sugar for `conn.check_request(conn.send_request_checked(req))`
+    ///
+    /// This method is useful as well in place of code sending a void request
+    /// and flushing the connection right after. Checking the request effectively
+    /// flushes the connection, but in addition reports possible protocol errors
+    /// at the calling site instead of reporting them through the event loop.
+    ///
+    /// # Example
+    /// ```no_run
+    /// # use xcb::x;
+    /// # fn main() -> xcb::Result<()> {
+    /// #   let (conn, screen_num) = xcb::Connection::connect(None)?;
+    /// #   let window: x::Window = conn.generate_id();
+    ///     conn.send_and_check_request(&x::MapWindow { window })?;
+    /// #   Ok(())
+    /// # }
+    /// ```
+    pub fn send_and_check_request<R>(&self, req: &R) -> ProtocolResult<()>
+    where
+        R: RequestWithoutReply,
+    {
+        self.check_request(self.send_request_checked(req))
+    }
+
+    /// Gets the reply of a previous request, or an error if one occurred.
+    ///
+    /// This is blocking; it does not return until the reply has been received. For the non-blocking
+    /// version, see [`poll_for_reply`].
     ///
     /// # Example
     /// ```no_run
@@ -1475,31 +1531,25 @@ impl Connection {
     /// #   Ok(())
     /// # }
     /// ```
+    ///
+    /// [`poll_for_reply`]: Self::poll_for_reply
     pub fn wait_for_reply<C>(&self, cookie: C) -> Result<C::Reply>
     where
         C: CookieWithReplyChecked,
     {
         unsafe {
-            let mut error: *mut xcb_generic_error_t = std::ptr::null_mut();
+            let mut error: *mut xcb_generic_error_t = ptr::null_mut();
             let reply = xcb_wait_for_reply64(self.c, cookie.sequence(), &mut error as *mut _);
-            match (reply.is_null(), error.is_null()) {
-                (true, true) => {
-                    self.has_error()?;
-                    unreachable!("xcb_wait_for_reply64 returned null without I/O error");
-                }
-                (true, false) => {
-                    let error = error::resolve_error(error, &self.ext_data);
-                    Err(error.into())
-                }
-                (false, true) => Ok(C::Reply::from_raw(reply as *const u8)),
-                (false, false) => unreachable!("xcb_wait_for_reply64 returned two pointers"),
-            }
+            self.handle_reply_checked::<C>(reply, error)
         }
     }
 
     /// Get the reply of a previous unchecked request.
     ///
-    /// If an error occured, `None` is returned and the error will be delivered to the event loop.
+    /// If an error occurred, `None` is returned and the error will be delivered to the event loop.
+    ///
+    /// This is blocking; it does not return until the reply has been received. For the non-blocking
+    /// version, see [`poll_for_reply_unchecked`].
     ///
     /// # Example
     /// ```no_run
@@ -1516,17 +1566,184 @@ impl Connection {
     /// #   Ok(())
     /// # }
     /// ```
+    ///
+    /// [`poll_for_reply_unchecked`]: Self::poll_for_reply_unchecked
     pub fn wait_for_reply_unchecked<C>(&self, cookie: C) -> ConnResult<Option<C::Reply>>
     where
         C: CookieWithReplyUnchecked,
     {
         unsafe {
             let reply = xcb_wait_for_reply64(self.c, cookie.sequence(), ptr::null_mut());
-            if reply.is_null() {
-                self.has_error()?;
-                Ok(None)
-            } else {
-                Ok(Some(C::Reply::from_raw(reply as *const u8)))
+            self.handle_reply_unchecked::<C>(reply)
+        }
+    }
+
+    /// Gets the reply of a previous request if it has been received, or an error if one occurred.
+    ///
+    /// This is non-blocking; if no reply has been received yet, it returns [`None`]. For the
+    /// blocking version, see [`wait_for_reply`].
+    ///
+    /// # Examples
+    /// ```no_run
+    /// # use xcb::x;
+    /// # fn main() -> xcb::Result<()> {
+    /// #     let (conn, screen_num) = xcb::Connection::connect(None)?;
+    /// let (wm_protocols_cookie, wm_name_cookie) = (
+    ///     conn.send_request(&x::InternAtom {
+    ///         only_if_exists: true,
+    ///         name: b"WM_PROTOCOLS",
+    ///     }),
+    ///     conn.send_request(&x::InternAtom {
+    ///         only_if_exists: true,
+    ///         name: b"WM_NAME",
+    ///     }),
+    /// );
+    /// let (wm_protocols_atom, wm_name_atom) = {
+    ///     let (
+    ///         mut wm_protocols_atom,
+    ///         mut wm_name_atom,
+    ///     ) = (None, None);
+    ///
+    ///     loop {
+    ///         // If `wm_protocols_atom` is yet to be received, poll for it.
+    ///         if wm_protocols_atom.is_none() {
+    ///             wm_protocols_atom = conn
+    ///                 .poll_for_reply(&wm_protocols_cookie)
+    ///                 .transpose()?
+    ///                 .map(|reply| reply.atom());
+    ///         }
+    ///         // If `wm_name_atom` is yet to be received, poll for it.
+    ///         if wm_name_atom.is_none() {
+    ///             wm_name_atom = conn
+    ///                 .poll_for_reply(&wm_name_cookie)
+    ///                 .transpose()?
+    ///                 .map(|reply| reply.atom());
+    ///         }
+    ///
+    ///         // If both `wm_protocols_atom` and `wm_name_atom` have been
+    ///         // received, break from the loop.
+    ///         if let (
+    ///             Some(wm_protocols_atom),
+    ///             Some(wm_name_atom),
+    ///         ) = (wm_protocols_atom, wm_name_atom) {
+    ///             break (wm_protocols_atom, wm_name_atom);
+    ///         }
+    ///     }
+    /// };
+    /// #     Ok(())
+    /// # }
+    /// ```
+    ///
+    /// [`wait_for_reply`]: Self::wait_for_reply
+    pub fn poll_for_reply<C>(&self, cookie: &C) -> Option<Result<C::Reply>>
+    where
+        C: CookieWithReplyChecked,
+    {
+        unsafe {
+            let mut error: *mut xcb_generic_error_t = ptr::null_mut();
+            let mut reply: *mut c_void = ptr::null_mut();
+
+            let received = xcb_poll_for_reply64(
+                self.c,
+                cookie.sequence(),
+                &mut reply as *mut _,
+                &mut error as *mut _,
+            );
+
+            match received {
+                0 => None,
+                1 => Some(self.handle_reply_checked::<C>(reply, error)),
+                _ => panic!("unexpected return value from xcb_poll_for_reply64"),
+            }
+        }
+    }
+
+    /// Gets the reply of a previous unchecked request if it has been received.
+    ///
+    /// If an error occurred, [`None`] is returned and the error is delivered to the event loop.
+    ///
+    /// This is non-blocking; if no reply has been received yet, it returns
+    /// <code>[Some]\([None])</code>. For the blocking version, see [`wait_for_reply_unchecked`].
+    ///
+    /// # Examples
+    /// ```no_run
+    /// # use xcb::x;
+    /// # fn main() -> xcb::Result<()> {
+    /// #     let (conn, screen_num) = xcb::Connection::connect(None)?;
+    /// let (wm_protocols_cookie, wm_name_cookie) = (
+    ///     conn.send_request_unchecked(&x::InternAtom {
+    ///         only_if_exists: true,
+    ///         name: b"WM_PROTOCOLS",
+    ///     }),
+    ///     conn.send_request_unchecked(&x::InternAtom {
+    ///         only_if_exists: true,
+    ///         name: b"WM_NAME",
+    ///     }),
+    /// );
+    /// let (wm_protocols_atom, wm_name_atom) = {
+    ///     let (
+    ///         mut wm_protocols_atom,
+    ///         mut wm_name_atom,
+    ///     ) = (Some(None), Some(None));
+    ///
+    ///     loop {
+    ///         // If `wm_protocols_atom` is yet to be received, poll for it.
+    ///         if let Some(None) = wm_protocols_atom {
+    ///             wm_protocols_atom = conn
+    ///                 // connection error may happen
+    ///                 .poll_for_reply_unchecked(&wm_protocols_cookie)
+    ///                 .transpose()?
+    ///                 .map(|result| result.map(|reply| reply.atom()));
+    ///         }
+    ///         // If `wm_name_atom` is yet to be received, poll for it.
+    ///         if let Some(None) = wm_name_atom {
+    ///             wm_name_atom = conn
+    ///                 // connection error may happen
+    ///                 .poll_for_reply_unchecked(&wm_name_cookie)
+    ///                 .transpose()?
+    ///                 .map(|result| result.map(|reply| reply.atom()));
+    ///         }
+    ///
+    ///         match (wm_protocols_atom, wm_name_atom) {
+    ///             // If either `wm_protocols_atom` or `wm_name_atom` hasn't
+    ///             // been received, continue the loop.
+    ///             (Some(None), _) | (_, Some(None)) => continue,
+    ///
+    ///             // Otherwise, if both have been received, break from the
+    ///             // loop.
+    ///             (
+    ///                 wm_protocols_atom,
+    ///                 wm_name_atom,
+    ///             ) => break (
+    ///                 wm_protocols_atom.flatten(),
+    ///                 wm_name_atom.flatten(),
+    ///             ),
+    ///         }
+    ///     }
+    /// };
+    /// #     Ok(())
+    /// # }
+    /// ```
+    ///
+    /// [`wait_for_reply_unchecked`]: Self::wait_for_reply_unchecked
+    pub fn poll_for_reply_unchecked<C>(&self, cookie: &C) -> Option<ConnResult<Option<C::Reply>>>
+    where
+        C: CookieWithReplyUnchecked,
+    {
+        unsafe {
+            let mut reply: *mut c_void = ptr::null_mut();
+
+            let received = xcb_poll_for_reply64(
+                self.c,
+                cookie.sequence(),
+                &mut reply as *mut _,
+                ptr::null_mut(),
+            );
+
+            match received {
+                0 => None,
+                1 => Some(self.handle_reply_unchecked::<C>(reply)),
+                _ => panic!("unexpected return value from xcb_poll_for_reply64"),
             }
         }
     }
@@ -1552,12 +1769,84 @@ impl Connection {
     }
 }
 
+impl Connection {
+    unsafe fn handle_wait_for_event(&self, ev: *mut xcb_generic_event_t) -> Result<Event> {
+        if ev.is_null() {
+            self.has_error()?;
+            panic!("xcb_wait_for_event returned null with I/O error");
+        } else if is_error(ev) {
+            Err(error::resolve_error(ev as *mut _, &self.ext_data).into())
+        } else {
+            Ok(event::resolve_event(ev, &self.ext_data))
+        }
+    }
+
+    unsafe fn handle_poll_for_event(&self, ev: *mut xcb_generic_event_t) -> Result<Option<Event>> {
+        if ev.is_null() {
+            self.has_error()?;
+            Ok(None)
+        } else if is_error(ev) {
+            Err(error::resolve_error(ev as *mut _, &self.ext_data).into())
+        } else {
+            Ok(Some(event::resolve_event(ev, &self.ext_data)))
+        }
+    }
+
+    unsafe fn handle_reply_checked<C>(
+        &self,
+        reply: *mut c_void,
+        error: *mut xcb_generic_error_t,
+    ) -> Result<C::Reply>
+    where
+        C: CookieWithReplyChecked,
+    {
+        match (reply.is_null(), error.is_null()) {
+            (true, true) => {
+                self.has_error()?;
+                unreachable!("xcb_wait_for_reply64 returned null without I/O error");
+            }
+            (true, false) => {
+                let error = error::resolve_error(error, &self.ext_data);
+                Err(error.into())
+            }
+            (false, true) => Ok(C::Reply::from_raw(reply as *const u8)),
+            (false, false) => unreachable!("xcb_wait_for_reply64 returned two pointers"),
+        }
+    }
+
+    unsafe fn handle_reply_unchecked<C>(&self, reply: *mut c_void) -> ConnResult<Option<C::Reply>>
+    where
+        C: CookieWithReplyUnchecked,
+    {
+        if reply.is_null() {
+            self.has_error()?;
+            Ok(None)
+        } else {
+            Ok(Some(C::Reply::from_raw(reply as *const u8)))
+        }
+    }
+}
+
+impl AsRef<Connection> for Connection {
+    fn as_ref(&self) -> &Connection {
+        self
+    }
+}
+
 impl AsRawFd for Connection {
     fn as_raw_fd(&self) -> RawFd {
         unsafe { xcb_get_file_descriptor(self.c) }
     }
 }
 
+// SAFETY: We provide a valid xcb_connection_t that is valid for as long as required by the trait.
+#[cfg(feature = "as-raw-xcb-connection")]
+unsafe impl as_raw_xcb_connection::AsRawXcbConnection for Connection {
+    fn as_raw_xcb_connection(&self) -> *mut as_raw_xcb_connection::xcb_connection_t {
+        self.get_raw_conn().cast()
+    }
+}
+
 impl Drop for Connection {
     fn drop(&mut self) {
         #[cfg(feature = "debug_atom_names")]
@@ -1597,30 +1886,28 @@ mod dan {
     pub(crate) static mut DAN_CONN: *mut xcb_connection_t = ptr::null_mut();
 
     impl fmt::Debug for x::Atom {
+        #[allow(clippy::print_in_format_impl)]
         fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-            let mut write = |name: &str| -> fmt::Result {
-                f.write_fmt(format_args!("Atom(\"{}\" ; {})", name, self.resource_id()))
-            };
-            if let Some(name) = x::predefined_atom_name(*self) {
-                write(name)
-            } else {
-                let conn = unsafe { Connection::from_raw_conn(DAN_CONN) };
-
-                let cookie = conn.send_request(&x::GetAtomName { atom: *self });
-                let reply = conn.wait_for_reply(cookie).map_err(|err| {
-                    eprintln!(
-                        "Error during fmt::Debug of x::Atom (fetching atom name): {:#?}",
-                        err
-                    );
-                    fmt::Error
-                })?;
-
-                let name = reply.name().to_utf8();
-                write(&name)?;
-
-                mem::forget(conn);
-                Ok(())
+            if self.resource_id() == 0 {
+                return f.write_str("ATOM_NONE");
             }
+
+            let conn = unsafe { Connection::from_raw_conn(DAN_CONN) };
+
+            let cookie = conn.send_request(&x::GetAtomName { atom: *self });
+            let reply = conn.wait_for_reply(cookie).map_err(|err| {
+                eprintln!(
+                    "Error during fmt::Debug of x::Atom (fetching atom name): {:#?}",
+                    err
+                );
+                fmt::Error
+            })?;
+
+            let name = reply.name().to_utf8();
+            f.write_fmt(format_args!("Atom(\"{}\" ; {})", name, self.resource_id()))?;
+
+            mem::forget(conn);
+            Ok(())
         }
     }
 }
@@ -1663,6 +1950,17 @@ pub(crate) fn align_pad(base: usize, align: usize) -> usize {
     (-base & (align - 1)) as usize
 }
 
+/// Get a value from a pointer, using ptr::copy_nonoverlapping to avoid alignment issues
+///
+/// # Safety
+/// The pointer must point to a valid `T` value
+#[inline]
+pub(crate) unsafe fn value_from_ptr<T>(ptr: *const T) -> T {
+    let mut val = mem::MaybeUninit::<T>::uninit();
+    ptr::copy_nonoverlapping(ptr, val.as_mut_ptr(), 1);
+    val.assume_init()
+}
+
 #[test]
 fn test_align_pad() {
     // align 1
