@@ -118,7 +118,7 @@
 //!         data: b"My XCB Window",
 //!     });
 //!     // And check for success again
-//!     conn.check_request(cookie);
+//!     conn.check_request(cookie)?;
 //!
 //!     // We now show ("map" in X terminology) the window.
 //!     // This time we do not check for success, so we discard the cookie.
@@ -354,6 +354,8 @@ mod xproto {
     /// `GRAB_ANY` can be used in various requests such as `GrabKey`, `UngrabKey`, `xinput::GrabDeviceKey`...
     pub const GRAB_ANY: Keycode = 0;
 
+    pub const ATOM_ANY: Atom = Atom { res_id: 0 };
+
     /// Trait for element in a property list
     ///
     /// In events (e.g. `GetProperty::value`), it allows to assert that the format
@@ -388,10 +390,94 @@ mod xproto {
 
     include!(concat!(env!("OUT_DIR"), "/xproto.rs"));
 }
+
+/// An helper macro that generate a struct of atoms.
+///
+/// The struct provide a constructor `intern_all` that takes a `Connection` as parameter,
+/// interns all the atoms and return `xcb::Result<[struct name]>`.
+/// `intern_all` takes advantage of XCB asynchronous design by sending all the
+/// [`x::InternAtom`] requests before starting to wait for the first reply.
+/// Fields that refer to atoms not existing in the server are set to `x::ATOM_NONE`
+/// (i.e. `only_if_exists` is always set to `true`).
+///
+/// Both the struct and each field can receive visibility attributes.
+///
+/// # Example
+/// ```no_run
+/// # use xcb::x;
+/// xcb::atoms_struct! {
+///     #[derive(Copy, Clone, Debug)]
+///     pub(crate) struct Atoms {
+///         pub wm_protocols    => b"WM_PROTOCOLS",
+///         pub wm_del_window   => b"WM_DELETE_WINDOW",
+///         /// Supported EWMH hints
+///         pub net_supported  => b"_NET_SUPPORTED",
+///
+///         // You can also explicitly set the `only_if_exists` argument when interning
+///         // each atom with the following syntax (the default is `true`):
+///         pub custom_atom    => b"MY_CUSTOM_ATOM" only_if_exists = false,
+///     }
+/// }
+///
+/// fn main() -> xcb::Result<()> {
+/// #   let (conn, screen_num) = xcb::Connection::connect(None)?;
+/// #   let window = conn.generate_id();
+///     // ...
+///     let atoms = Atoms::intern_all(&conn)?;
+///
+///     conn.check_request(conn.send_request_checked(&x::ChangeProperty {
+///         mode: x::PropMode::Replace,
+///         window,
+///         property: atoms.wm_protocols,
+///         r#type: x::ATOM_ATOM,
+///         data: &[atoms.wm_del_window],
+///     }))?;
+///     // ...
+/// #   Ok(())
+/// }
+/// ```
+#[macro_export]
+macro_rules! atoms_struct {
+    (
+        $(#[$outer:meta])*
+        $vis:vis struct $Atoms:ident {
+            $(
+                $(#[$fmeta:meta])* $fvis:vis $field:ident => $name:tt $( only_if_exists = $only_if_exists:expr)?,
+            )*
+        }
+    ) => {
+        $(#[$outer])*
+        $vis struct $Atoms {
+            $($(#[$fmeta])* $fvis $field: xcb::x::Atom,)*
+        }
+        impl $Atoms {
+            #[allow(dead_code)]
+            pub fn intern_all(conn: &xcb::Connection) -> xcb::Result<$Atoms> {
+                $(
+                    #[allow(unused_assignments)]
+                    let mut only_if_exists = true;
+                    $( only_if_exists = $only_if_exists; )?
+                    let $field = conn.send_request(&xcb::x::InternAtom {
+                        only_if_exists,
+                        name: $name,
+                    });
+                )*
+                $(
+                    let $field = conn.wait_for_reply($field)?.atom();
+                )*
+                Ok($Atoms {
+                    $($field,)*
+                })
+            }
+        }
+    };
+}
+
 pub mod bigreq {
     //! The `BIG-REQUESTS` extension.
     include!(concat!(env!("OUT_DIR"), "/bigreq.rs"));
 }
+
 pub mod xc_misc {
     //! The `XC-MISC` extension.
     include!(concat!(env!("OUT_DIR"), "/xc_misc.rs"));
@@ -535,6 +621,50 @@ pub mod randr {
     #![allow(clippy::unit_arg)]
     #![allow(clippy::too_many_arguments)]
     include!(concat!(env!("OUT_DIR"), "/randr.rs"));
+
+    /// For information on what the various properties mean, see the [RandR specification][randr-spec]
+    ///
+    /// [randr-spec]: https://cgit.freedesktop.org/xorg/proto/randrproto/tree/randrproto.txt#n1860
+    mod property {
+        #[doc(alias = "RR_PROPERTY_BACKLIGHT")]
+        pub const BACKLIGHT: &str = "Backlight";
+
+        #[doc(alias = "RR_PROPERTY_CLONE_LIST")]
+        pub const CLONE_LIST: &str = "CloneList";
+
+        #[doc(alias = "RR_PROPERTY_COMPATIBILITY_LIST")]
+        pub const COMPATIBILITY_LIST: &str = "CompatibilityList";
+
+        #[doc(alias = "RR_PROPERTY_CONNECTOR_NUMBER")]
+        pub const CONNECTOR_NUMBER: &str = "ConnectorNumber";
+
+        #[doc(alias = "RR_PROPERTY_CONNECTOR_TYPE")]
+        pub const CONNECTOR_TYPE: &str = "ConnectorType";
+
+        #[doc(alias = "RR_PROPERTY_RANDR_EDID")]
+        pub const EDID: &str = "EDID";
+
+        #[doc(alias = "RR_PROPERTY_SIGNAL_FORMAT")]
+        pub const SIGNAL_FORMAT: &str = "SignalFormat";
+
+        #[doc(alias = "RR_PROPERTY_SIGNAL_PROPERTIES")]
+        pub const SIGNAL_PROPERTIES: &str = "SignalProperties";
+
+        #[doc(alias = "RR_PROPERTY_BORDER")]
+        pub const BORDER: &str = "Border";
+
+        #[doc(alias = "RR_PROPERTY_BORDER_DIMENSIONS")]
+        pub const BORDER_DIMENSIONS: &str = "BorderDimensions";
+
+        #[doc(alias = "RR_PROPERTY_GUID")]
+        pub const GUID: &str = "GUID";
+
+        #[doc(alias = "RR_PROPERTY_RANDR_TILE")]
+        pub const TILE: &str = "TILE";
+
+        #[doc(alias = "RR_PROPERTY_NON_DESKTOP")]
+        pub const NON_DESKTOP: &str = "non-desktop";
+    }
 }
 
 #[cfg(feature = "record")]
@@ -677,6 +807,7 @@ pub mod xkb {
     //!
     //! Accessible with the `xkb` cargo feature.
     #![allow(unused_variables)]
+    #![allow(clippy::let_unit_value)]
     #![allow(clippy::unit_arg)]
     #![allow(clippy::too_many_arguments)]
     include!(concat!(env!("OUT_DIR"), "/xkb.rs"));
